@@ -7,6 +7,10 @@ use Term::ReadPassword;
 use File::Find;
 use DBI;
 use Data::Dumper;
+use Fcntl qw(:DEFAULT :flock);
+use File::Temp qw(tempfile);
+use Text::Iconv;
+use Carp;
 
 ################################################################################
 
@@ -46,7 +50,7 @@ sub _cd {
 sub _local_exec {
 
 	my $line = $_[0];	
-	$line =~ s{\-p\".*?\"}{-p********};
+	$line =~ s{\-p\".*?\"}{-p********}g;
 
 	print STDERR " {$line";
 
@@ -87,9 +91,6 @@ sub _master_exec {
 sub restore_local_libs {
 	my ($path) = @_;
 	$path ||= $ARGV [0];
-	_log ("Removing libs...");
-	_local_exec ("rm -rf lib/Content/*");
-	_local_exec ("rm -rf lib/Presentation/*");
 	_log ("Unpacking $path...");
 	_local_exec ("tar xzfm $path");
 	_local_exec ("chmod -R a+rwx lib/*");
@@ -104,8 +105,8 @@ sub restore_local_i {
 
 	my $local_preconf = _read_local_preconf ();
 
-	_log ("Removing i...");
-	_local_exec ("rm -rf docroot/i/*");
+#	_log ("Removing i...");
+#	_local_exec ("rm -rf docroot/i/*");
 	_log ("Unpacking $path...");
 	_local_exec ("tar xzfm $path");
 	_local_exec ("chmod -R a+rwx docroot/i/*");
@@ -142,7 +143,7 @@ sub restore {
 
 sub restore_local {
 
-	my ($time) = @_;
+	my ($time, $skip_libs) = @_;
 	$time ||= $ARGV [0];	
 	$time ||= readlink 'snapshots/latest.tar.gz';
 
@@ -161,11 +162,13 @@ sub restore_local {
 
 	my $local_preconf = _read_local_preconf ();
 	my $local_conf    = _read_local_conf    ();
-
-	my $lib_path = 'lib/' . $local_conf -> {application_name} . '.' . $time . '.tar.gz';
-	restore_local_libs ($lib_path);
-	_log ("Removing $lib_path...");
-	_local_exec ("rm $lib_path");
+	
+	unless ($skip_libs) {
+		my $lib_path = 'lib/' . $local_conf -> {application_name} . '.' . $time . '.tar.gz';
+		restore_local_libs ($lib_path);
+		_log ("Removing $lib_path...");
+		_local_exec ("rm $lib_path");
+	}
 	
 	if ($local_preconf -> {master_server} -> {static} eq 'none') {
 		_log ("RESTORING STATIC FILES ON LOCAL SERVER IS SKIPPED.");
@@ -320,7 +323,7 @@ sub sync_down {
 	_log ("Copying $snapshot_path from master...");
 	_cp_from_master ($local_preconf, $local_preconf -> {master_server} -> {path} . '/' . $snapshot_path, 'snapshots/');
 	
-	restore_local ($snapshot_path);
+	restore_local ($snapshot_path, $local_preconf -> {master_server} -> {skip_libs});
 
 	my $timing = ', ' . (time - $time) . ' s elapsed.';
 
@@ -419,8 +422,11 @@ sub backup_master {
 	my $db_path = backup_master_db ($time);
 	$db_path =~ s{$master_path\/}{};
 	
-	my $libs_path = backup_master_libs ($time);
-	$libs_path =~ s{$master_path\/}{};	
+	my $libs_path = '';
+	unless ($local_preconf -> {master_server} -> {skip_libs}) {
+		$libs_path = backup_master_libs ($time);
+		$libs_path =~ s{$master_path\/}{};	
+	}	
 	
 	my $i_path = backup_master_i ($time);
 	$i_path =~ s{$master_path\/}{};	
@@ -434,8 +440,10 @@ sub backup_master {
 	_log ("Removing $db_path...");
 	_master_exec ($local_preconf, "cd $master_path; rm $db_path");
 	
-	_log ("Removing $libs_path...");
-	_master_exec ($local_preconf, "cd $master_path; rm $libs_path");
+	unless ($local_preconf -> {master_server} -> {skip_libs}) {
+		_log ("Removing $libs_path...");
+		_master_exec ($local_preconf, "cd $master_path; rm $libs_path");
+	}
 	
 	if ($i_path) {
 		_log ("Removing $i_path...");
@@ -458,18 +466,8 @@ sub backup_local_db {
 	my $path = _db_path ($local_preconf -> {db_name}, $time);
 	_log ("Backing up db $$local_preconf{db_name} on local server...");
 	
-	_log ("Creating ${path}_...");
-	_local_exec ("mysqldump --add-drop-table -u$$local_preconf{db_user} -p\"$$local_preconf{db_password}\" $$local_preconf{db_name} > ${path}_");
-	
-	_log ("Cleaning up: ${path}_ -> $path...");
-	_local_exec ("cat ${path}_ | perl -ne \"s{ TYPE=\\w+}{}; print\" > $path");
-	
-	_log ("Deleting ${path}_...");
-	_local_exec ("rm ${path}_");
-
-	_log ("Gzipping $path...");
-	_local_exec ("gzip $path");
-	
+	_local_exec ("mysqldump --add-drop-table -Keq -u$$local_preconf{db_user} -p\"$$local_preconf{db_password}\" $$local_preconf{db_name} | gzip > $path.gz");
+		
 	_log ("DB backup complete");
 	return "$path.gz";
 		
@@ -489,22 +487,29 @@ sub backup_master_db {
 	my $path = $local_preconf -> {master_server} -> {path} . '/' . _db_path ($local_preconf -> {db_name}, $time);
 	_log ("Backing up db $$master_preconf{db_name} on master server...");
 	
-	_log ("Creating ${path}_...");
-	_master_exec ($local_preconf, "mysqldump --add-drop-table -u$$master_preconf{db_user} -p\"$$master_preconf{db_password}\" $$master_preconf{db_name} > ${path}_");
-
-	my $table_filter = '';
-	if (ref $local_preconf -> {master_server} -> {skip_tables} eq ARRAY) {			
-		foreach my $table (@{$local_preconf -> {master_server} -> {skip_tables}}) {
-			$table_filter .= qq{| grep -v "INSERT INTO $table"};
-		}	
+	_log ("Listing tables...");	
+	my $tables = _master_exec ($local_preconf, "mysql -u$$master_preconf{db_user} -p\"$$master_preconf{db_password}\" $$master_preconf{db_name} -se\"show tables\"");
+	
+	my %skip_tables = map {$_ => 1} @{$local_preconf -> {master_server} -> {skip_tables}};
+			
+	my @tables = split /\n/, $tables;
+	
+	map {s{\s}{}gsm} @tables;
+	
+	my $dump_cmd = '';
+	
+	foreach my $table (@tables) {
+			
+		my $opt = '-Keq';
+		$opt .= 'd' if $skip_tables {$table};
+				
+		$dump_cmd .= "mysqldump $opt --add-drop-table -u$$master_preconf{db_user} -p\"$$master_preconf{db_password}\" $$master_preconf{db_name} $table >> $path; ";
+			
 	}
-
-	_log ("Cleaning up: ${path}_ -> $path...");
-	_master_exec ($local_preconf, "cat ${path}_ $table_filter | perl -ne \"s{ TYPE=\\w+}{}; print\" > $path");
-
-	_log ("Deleting ${path}_...");
-	_master_exec ($local_preconf, "rm ${path}_");
-		
+	
+	_log ("Dumping database...");
+	_master_exec ($local_preconf, $dump_cmd);
+					
 	_log ("Gzipping $path...");
 	_master_exec ($local_preconf, "gzip $path");
 	
@@ -591,7 +596,7 @@ sub backup_master_i {
 	my $local_preconf = _read_local_preconf ();
 	my $local_conf = _read_local_conf ();
 
-	if ($preconf -> {master_server} -> {static} eq 'none' or $preconf -> {master_server} -> {static} eq 'rsync') {
+	if ($local_preconf -> {master_server} -> {static} eq 'none' or $local_preconf -> {master_server} -> {static} eq 'rsync') {
 		_log ("BACKING UP STATIC FILES ON MASTER SERVER IS SKIPPED.");
 		return '';
 	}
@@ -633,13 +638,19 @@ sub restore_master_libs {
 
 sub _read_master_conf {
 
-	my $local_preconf = _read_local_preconf ();
-	my $local_conf = _read_local_conf ();
+	unless ($MASTER_CONF) {
 
-	my $src = _master_exec ($local_preconf, 'cat ' . $local_preconf -> {master_server} -> {path} . "/lib/$$conf{application_name}/Config.pm");
-	undef $conf;
-	eval $src;
-	return $conf;
+		my $local_preconf = _read_local_preconf ();
+		my $local_conf = _read_local_conf ();
+
+		my $src = _master_exec ($local_preconf, 'cat ' . $local_preconf -> {master_server} -> {path} . "/lib/$$conf{application_name}/Config.pm");
+		undef $conf;
+		eval $src;
+		$MASTER_CONF = $conf;
+		
+	}
+
+	return $MASTER_CONF;
 
 }
 
@@ -647,11 +658,35 @@ sub _read_master_conf {
 
 sub _read_master_preconf {
 
-	my $local_preconf = _read_local_preconf ();
-	my $local_conf = _read_local_conf ();
+	unless ($MASTER_PRECONF) {
 
-	my $src = _master_exec ($local_preconf, 'cat ' . $local_preconf -> {master_server} -> {path} . "/conf/httpd.conf");
-	return _decrypt_preconf ($src);
+		my $local_preconf = _read_local_preconf ();
+		my $local_conf = _read_local_conf ();
+
+		my $src = _master_exec ($local_preconf, 'cat ' . $local_preconf -> {master_server} -> {path} . "/conf/httpd.conf");
+		$MASTER_PRECONF = _decrypt_preconf ($src);
+
+	}
+
+	return $MASTER_PRECONF;
+	
+}
+
+################################################################################
+
+sub _read_local_preconf {
+
+	unless ($LOCAL_PRECONF) {
+	
+		_cd ();
+
+		-f 'conf/httpd.conf' or die "ERROR: httpd.conf not found. Please, first chdir to the webapp directory.\n";
+		my $src = `cat conf/httpd.conf`;
+		$LOCAL_PRECONF = _decrypt_preconf ($src);
+		
+	}
+
+	return $LOCAL_PRECONF;
 	
 }
 
@@ -659,14 +694,21 @@ sub _read_master_preconf {
 
 sub _read_local_conf {
 	
-	_cd ();
+	unless ($LOCAL_CONF) {
+
+		_cd ();
+
+		opendir (DIR, 'lib') || die "can't opendir lib: $!";
+		my ($appname) = grep {(-d "lib/$_") && ($_ !~ /\./) } readdir (DIR);
+		closedir DIR;
+		do "lib/$appname/Config.pm";
+		$conf -> {application_name} = $appname;
+		
+		$LOCAL_CONF = $conf;
+
+	}
 	
-	opendir (DIR, 'lib') || die "can't opendir lib: $!";
-	my ($appname) = grep {(-d "lib/$_") && ($_ !~ /\./) } readdir (DIR);
-	closedir DIR;
-	do "lib/$appname/Config.pm";
-	$conf -> {application_name} = $appname;
-	return $conf;	
+	return $LOCAL_CONF;
 	
 }
 
@@ -705,15 +747,11 @@ sub _decrypt_preconf {
 }
 
 ################################################################################
-
-sub _read_local_preconf {
-
-	_cd ();
-
-	-f 'conf/httpd.conf' or die "ERROR: httpd.conf not found. Please, first chdir to the webapp directory.\n";
-	my $src = `cat conf/httpd.conf`;
-	return _decrypt_preconf ($src);
-	
+sub _require_svn {
+    eval 'require SVN::Client;';
+    if ($@) {
+	die "SVN::Client not found. You should install perl binding from http://subversion.tigris.org/. Debian Linux users should install libsvn-core-perl package\n";
+    }
 }
 
 ################################################################################
@@ -724,12 +762,14 @@ sub create {
 	
 	$path =~ s{Install\.pm}{static/sample.tar.gz.pm};
 	
-	my $term = new Term::ReadLine 'Zanas application installation';
+	our $term = new Term::ReadLine 'Zanas application installation';
 	
 	my ($appname, $appname_uc, $instpath, $db, $user, $password, $group, $admin_user, $admin_password, $dbh);
 	
+	my ($svn_ctx, $svn_server, $application_src, $application_dst, $appname_src, $appname_src_uc);
+	our ($is_svn_error, $svn_error_str);
+
 	while (1) {
-	
 		while (1) {
 			$appname = $term -> readline ('Application name (lowercase): ');
 			last if $appname =~ /[a-z_]+/
@@ -741,6 +781,11 @@ sub create {
 			my $default_instpath = $^O eq 'MSWin32' ? "G:\\do_work\\$appname" : "/var/projects/$appname";
 			$instpath = $term -> readline ("Installation path [$default_instpath]: ");
 			$instpath = $default_instpath if $instpath eq '';
+			if (-d $instpath) {
+				print "Installation path exists.\n";
+				next;
+			}
+
 			last if $instpath =~ /[\w\/]+/
 		}
 
@@ -774,14 +819,83 @@ sub create {
 		}
 
 		$dbh -> {RaiseError} = 1;
+
+		print "\nApplication source\n\tpub_sample - from Zanas::Install package\n\tsvn://svn.zanas.ru/pub_sample - public web site template from svn\n\tsvn://svn.zanas.ru/sample - web application template from svn (stripped public web site template)\n\tsvn://.... - your own source\n";
+		while (1) {
+			$application_src = $term -> readline ('[svn://svn.zanas.ru/pub_sample]: ');
+			$application_src = 'svn://svn.zanas.ru/pub_sample' unless $application_src;
+			if ($application_src eq 'pub_sample') {
+				last;
+			} else {
+				_require_svn ();
+
+				$SVN::Error::handler = \&svn_error_handler;
+				$svn_ctx = new SVN::Client(
+					auth => [
+						SVN::Client::get_simple_provider(),
+						SVN::Client::get_simple_prompt_provider(\&svn_simple_prompt, 10)
+					]);
+				my $svn_props = $svn_ctx -> proplist ($application_src . '/trunk', 'HEAD', 0);
+				last unless ($is_svn_error);
+				$is_svn_error = 0;
+				print "Bad SVN repository url. Try again\n";
+			}
+		}
+		print "\n";
+		
+		$appname_src = 'sample';
+		if ($application_src !~ 'sample') {
+			$application_src =~ m{.*\://[^/]+/([^/]+)};
+			my $appname_src_dflt = $1;
+			while (1) {
+				$appname_src = $term -> readline ("Source application name (lowercase): [$appname_src_dflt]: ");
+				$appname_src = $appname_src_dflt unless $appname_src;
+
+				last if $appname_src =~ /[a-z_]+/
+			}
+		}
+
+# source - svn repository
+		unless ($application_src eq 'pub_sample') {
+			$application_src =~ m{(.*\://[^/]+)};
+			my $svn_server_url = $1;
+			my $default_url = "$svn_server_url/$appname";
+			if ($appname eq $appname_src) {
+				my $default_url = 'same';
+			}
+			print "\nApplication's SVN repository destination\n\tnone - don't create\n\tsame - don't create branch, leave source repository url (only when application name equal source application name)\n\turl - create new branch at url location\n";
+			while (1) {
+				$application_dst = $term -> readline ("[$default_url]: ");
+				$application_dst = $default_url unless $application_dst;
+				
+				if ($application_dst eq 'same' && $appname ne $appname_src) {
+					print "Can't use the same repository url: application name not equal source application name. Try again\n";
+				} else {
+					unless ($application_dst eq 'none' || $application_dst eq 'same') {
+						my $svn_props = $svn_ctx -> proplist ($application_dst . '/trunk', 'HEAD', 0);
+						last if $is_svn_error == $SVN::Error::NODE_UNKNOWN_KIND;
+						print "Bad or busy SVN repository url. Try again\n";
+					} else {
+						last;
+					}
+				}
+			}
+		}
+		
+		
+		$appname_src_uc = uc $appname_src;	
 		
 		$password = random_password ();
 
 		print <<EOT;
+		
 			Application name:	$appname
 			User group:		$group
 			Database name:		$db
 			Database user:		$user
+			Application template source:	$application_src
+			SVN repository destination:	$application_dst
+			Source application name:	$appname_src_uc
 EOT
 			
 		my $ok = $term -> readline ("Everything in its right place? (yes/NO): ");
@@ -790,8 +904,6 @@ EOT
 		
 	}
 	
-	-d $instpath and die ("Can't proceed: installation path exists.\n");
-		
 	print "Creating database... ";
 		
 	$dbh -> do ("CREATE DATABASE $db");
@@ -816,26 +928,52 @@ EOT
 		eval 'require Archive::Tar;';
 		my $tar = Archive::Tar -> new ();
 		$tar -> read ($path, 1);
-		$tar -> extract ($tar -> list_files ());
+		if ($application_src eq 'pub_sample') {
+# static template
+			$tar -> extract ($tar -> list_files ());
+		} else {
+			$tar -> extract ([(grep {$_ =~ /^(conf)|(logs)|(snapshots)|(sql)/} @{$tar -> list_files ()})]);
+		}
+	} else {
+		if ($application_src eq 'pub_sample') {
+			`tar xzfm $path --directory=$instpath/`;
+		} else {
+			`tar xzfm $path --directory=$instpath/ conf logs snapshots sql`;
+		}
+	}
+	
+	if ($application_src ne 'pub_sample' && $application_dst ne 'none' && $application_dst ne 'same') {
+# Create branch
+		$svn_ctx -> mkdir ("$application_dst");
+		$svn_ctx -> mkdir ("$application_dst/tags");
+		$svn_ctx->copy ($application_src . '/trunk', 'HEAD', "$application_dst/trunk");
+		$svn_ctx->checkout("$application_dst/trunk", $instpath, 'HEAD', 1);
+	} elsif ($application_dst eq 'same' || $application_dst eq 'none') {
+		$svn_ctx->checkout("$application_src/trunk", $instpath, 'HEAD', 1);
+		if ($application_dst eq 'none') {
+# Delete SVN system files
+			find (sub {`rm -rf $File::Find::name` if $File::Find::name =~ /\.svn$/}, "$instpath");
+		}
 	}
-	else {
-		`tar xzfm $path --directory=$instpath/`;
-	}	
 	print "ok\n";
 
 	print "Renaming application files... ";
-	if ($^O eq 'MSWin32') {
-		rename "$instpath\\lib\\SAMPLE", "$instpath\\lib\\$appname_uc";
-		rename "$instpath\\lib\\SAMPLE.pm", "$instpath\\lib\\$appname_uc.pm";
+	if ($application_src ne 'pub_sample' && $application_dst ne 'none' && $application_dst ne 'same') {
+		$svn_ctx->move ("$instpath/lib/$appname_src_uc", 'HEAD', "$instpath/lib/$appname_uc", 1);
+		$svn_ctx->move ("$instpath/lib/$appname_src_uc.pm", 'HEAD', "$instpath/lib/$appname_uc.pm", 1);
+	} elsif ($appname_src ne $appname) {
+		if ($^O eq 'MSWin32') {
+			rename "$instpath\\lib\\$appname_src_uc", "$instpath\\lib\\$appname_uc";
+			rename "$instpath\\lib\\$appname_src_uc.pm", "$instpath\\lib\\$appname_uc.pm";
+		} else {
+			`mv $instpath/lib/$appname_src_uc $instpath/lib/$appname_uc`;
+			`mv $instpath/lib/$appname_src_uc.pm $instpath/lib/$appname_uc.pm`;
+		}	
 	}
-	else {
-		`mv $instpath/lib/SAMPLE $instpath/lib/$appname_uc`;
-		`mv $instpath/lib/SAMPLE.pm $instpath/lib/$appname_uc.pm`;
-	}	
 	print "ok\n";
 	
 	our %substitutions = (
-		SAMPLE => $appname_uc,
+		$appname_src_uc => $appname_uc,
 	);
 	
 	if ($^O eq 'MSWin32') {
@@ -846,11 +984,12 @@ EOT
 	}	
 
 	%substitutions = (
-		'/var/projects/sample' => $instpath,
-		'=sample' => '=' . $appname,
-		SAMPLE => $appname_uc,
-		"'do'" => "'$user'",
-		"'z'" => "'$password'",
+		"/var/projects/$appname_src" => $instpath,
+		'/var/vh/$appname_src' => $instpath,
+		"=$appname_src" => '=' . $appname,
+		$appname_src_uc => $appname_uc,
+		"db_user => 'do'" => "db_user => '$user'",
+		"db_password => 'z'" => "db_password => '$password'",
 	);
 	
 	if ($^O eq 'MSWin32') {
@@ -865,6 +1004,9 @@ EOT
 		`chmod -R a+w $instpath`;
 	}
 
+	if ($application_src ne 'pub_sample' && $application_dst ne 'none' && $application_dst ne 'same') {
+		$svn_ctx -> commit (["$instpath/lib", "$instpath/docroot"], 0);
+	}
 	print <<EOT;
 
 --------------------------------------------------------------------------------
@@ -890,6 +1032,234 @@ d.o.
 EOT
 
 }
+#'
+
+################################################################################
+
+sub svn_import {
+
+	_cd ();
+	-f 'conf/httpd.conf' or die "ERROR: httpd.conf not found. Please, first chdir to the webapp directory.\n";
+
+	my $instpath = `pwd`;
+	chomp ($instpath);
+	$instpath =~ m{([^/]+)$};
+	my $appname = $1;
+	my $svn_server_url = 'svn://svn.zanas.ru';
+
+
+	eval 'require SVN::Client;';
+	our ($is_svn_error, $svn_error_str);
+	$SVN::Error::handler = \&svn_error_handler;
+	my $svn_ctx = new SVN::Client(
+		auth => [
+			SVN::Client::get_simple_provider(),
+			SVN::Client::get_simple_prompt_provider(\&svn_simple_prompt, 10)
+		]);
+
+	our $term = new Term::ReadLine 'Zanas application installation';
+
+	while (1) {
+		$application_dst = $term -> readline ("Application's SVN repository destination [$svn_server_url/$appname]: ");
+		$application_dst = "$svn_server_url/$appname" unless $application_dst;
+
+		my $svn_props = $svn_ctx -> proplist ($application_dst . '/trunk', 'HEAD', 0);
+		last if $is_svn_error == $SVN::Error::NODE_UNKNOWN_KIND;
+		print "Bad or busy SVN repository url. Try again\n";
+	}
+		
+	backup_local ();
+
+	find (sub {`rm -rf $File::Find::name` if $File::Find::name =~ /\.svn$/}, "$instpath");
+	print "Importing application files... ";
+	$svn_ctx -> mkdir ("$application_dst");
+	$svn_ctx -> mkdir ("$application_dst/tags");
+	$svn_ctx -> mkdir ("$application_dst/trunk");
+	$svn_ctx -> mkdir ("$application_dst/trunk/docroot");
+	$svn_ctx -> mkdir ("$application_dst/trunk/docroot/i");
+
+	opendir (DIR, 'docroot/i') || die "can't opendir docroot/i: $!";
+	map {
+		unless ($_ =~ /upload/ || $_ eq '.' || $_ eq '..') {
+#			print "$_\n";
+			my $result = $svn_ctx -> import ("$instpath/docroot/i/$_", "$application_dst/trunk/docroot/i/$_", 0) ;
+			die "Error importing directory $_\n" unless ($result);
+		}
+	} readdir (DIR);
+	closedir (DIR);
+	
+	my $result = $svn_ctx -> import ("$instpath/lib", "$application_dst/trunk/lib", 0);
+	die "Error importing directory $instpath/lib\n" unless ($result);
+	mkdir "docroot_save";
+	opendir (DIR, 'docroot') || die "can't opendir docroot: $!";
+	map {
+		unless ($_ eq 'i' || $_ eq '.' || $_ eq '..') {
+		    `mv docroot/$_ docroot_save/$_`;
+		}
+	} readdir (DIR);
+	closedir (DIR);
+
+	if (-d "docroot/i/upload") {
+	    mkdir 'docroot_save/i'; 
+	    `mv docroot/i/upload docroot_save/i/`;
+	}
+        `rm -rf $instpath/docroot`;
+	`rm -rf "$instpath/lib"`;
+
+	$svn_ctx->checkout("$application_dst/trunk", "$instpath", 'HEAD', 1);
+
+	`cp -r docroot_save/* docroot/`;
+	`rm -rf docroot_save`;
+
+	print "ok\n";
+
+}
+
+################################################################################
+
+sub svn_commit {
+
+	_cd ();
+	-f 'conf/httpd.conf' or die "ERROR: httpd.conf not found. Please, first chdir to the webapp directory.\n";
+	our $term = new Term::ReadLine 'Zanas application installation';
+
+	
+	eval 'require SVN::Client;';
+	our ($is_svn_error, $svn_error_str);
+	$SVN::Error::handler = \&svn_error_handler;
+
+	my $svn_ctx = new SVN::Client(
+		auth => [
+			SVN::Client::get_simple_provider(),
+			SVN::Client::get_simple_prompt_provider(\&svn_simple_prompt, 10)
+		]);
+	my $tag;
+	my $url = $svn_ctx -> url_from_path('lib');
+	$url || die 'Application is unversioned.';
+	$url =~ s{/trunk/lib$}{};
+
+        my $tag = $term -> readline ('Tag version as testing? (yes/NO): ');
+                
+	my $cnt = 0;
+	for my $path ('lib', 'docroot/i') {
+#		print "path: $path\n";
+		$svn_ctx -> status ($path, 'HEAD', 
+			sub {
+				my ($file, $status) = @_; 
+				next if ($file =~ m{/i/upload});
+#				print "$file: ";
+				if ($status -> text_status() == $SVN::Wc::Status::unversioned) {
+					$cnt ++;
+#					print "unversioned\n";
+					$svn_ctx -> add ($file, 1);
+				} elsif ($status -> text_status() == $SVN::Wc::Status::obstructed) {
+					$cnt ++;
+#					print "obstructed\n";
+				} elsif ($status -> text_status() == $SVN::Wc::Status::missing) {
+					$cnt ++;
+#					print "missing\n";
+					$svn_ctx -> delete ($file, 1);
+				} elsif ($status -> text_status() == $SVN::Wc::Status::modified) {
+					$cnt ++;
+#					print "modified\n";
+				} elsif ($status -> text_status() == $SVN::Wc::Status::normal) {
+#					print "normal\n";
+				} else {
+#					print $status -> text_status() . "\n";
+					$cnt ++; # ,    
+				}
+			}, 
+	# recursive
+			1, 
+	# retrieve all entries
+			1, 
+	# contact the repository
+			1, 
+	# disregard default and svn:ignore property ignores
+			0);
+	}
+
+	my $result;
+	if ($cnt > 0) {
+	    our $log_message;
+	    $svn_ctx -> log_msg(\&solicit);
+            $result = $svn_ctx -> commit(['lib', 'docroot'], 0);
+	    unless ($result) {
+	        print "Commit error: $svn_error_str\n";
+		return;
+	    }
+	} else {
+	    print "Nothing commit\n";
+	}
+
+	if ($tag eq 'yes') {
+	    $svn_ctx -> log_msg(undef);
+	    my $revision = $result && $result->revision();
+	    if (!$revision || $revision == $SVN::Core::INVALID_REVNUM) {
+		my $dir = $svn_ctx->ls("$url", 'HEAD', 0);
+		$revision = $dir -> {trunk} -> created_rev();
+#	        $revision = ($svn_ctx->revprop_list("$url/trunk", 'HEAD')) [1];
+	    
+	    }
+	    print "Current revision: $url/trunk/lib: $revision\n";
+	    
+	    $is_svn_error = 0;
+            my $svn_props = $svn_ctx -> proplist ("$url/tags/testing_$revision", 'HEAD', 0);
+            if ($is_svn_error == 0) {
+		print "Current revision is already tagged as testing\n";
+	    } else {
+	        $svn_ctx->copy ("$url/trunk", $revision, "$url/tags/testing_$revision");
+	    }
+	}
+
+	print "ok\n";
+}
+
+################################################################################
+
+sub svn_update {
+	my $rev = shift;
+	
+	$rev ||= 'HEAD';
+	
+	_cd ();
+	-f 'conf/httpd.conf' or die "ERROR: httpd.conf not found. Please, first chdir to the webapp directory.\n";
+
+	eval 'require SVN::Client;';
+	our ($is_svn_error, $svn_error_str);
+	$SVN::Error::handler = \&svn_error_handler;
+	my $svn_ctx = new SVN::Client(
+		auth => [
+			SVN::Client::get_simple_provider(),
+			SVN::Client::get_simple_prompt_provider(\&svn_simple_prompt, 10)
+		]);
+
+	$svn_ctx -> update ('lib', $rev, 1);
+}
+
+################################################################################
+sub svn_simple_prompt {
+	my $cred = shift;
+	my $realm = shift;
+	my $default_username = shift;
+	my $may_save = shift;
+	my $pool = shift;
+
+	my $username = $term -> readline ("SVN repository user [$ENV{USER}]: ");
+	$username = $ENV{USER} unless ($username);
+	$cred->username($username);
+	my $password = $term -> readline ("SVN repository password: ");
+	$cred->password($password);
+}
+
+################################################################################
+sub svn_error_handler {
+	my $svn_error = shift;
+	$is_svn_error = $svn_error -> apr_err();
+	$svn_error_str = $svn_error -> strerror();
+#	print "$svn_error_str: $is_svn_error\n";
+	$svn_error->clear();
+}
 
 ################################################################################
 
@@ -926,6 +1296,66 @@ sub fix {
 
 }
 
+################################################################################
+
+
+sub solicit {
+	my ($msg, $file_path, $objects) = @_;
+#	print STDERR "solicit\n";
+	
+	my $info = "--This line, and those below, will be ignored--\n\n";
+	
+	File::Temp->safe_level(2);
+	my ( $tfh, $filename ) = tempfile( UNLINK => 1 );
+
+	unless ( $tfh and $filename ) {
+		die "\nno temporary file\n";
+	}
+
+	select( ( select($tfh), $|++ )[0] );
+
+	foreach my $obj (@$objects) {
+		my $flag = $obj->state_flags();
+		my $status = 
+			$flag & $SVN::Client::COMMIT_ITEM_ADD ? 'A' :
+			$flag & $SVN::Client::COMMIT_ITEM_DELETE ? 'D' :
+			$flag & $SVN::Client::COMMIT_ITEM_TEXT_MODS  ? 'M' : '';
+			
+		$info .= ($status || $flag) . "\t" . $obj -> path() . "\n";
+	}
+	print $tfh $info;
+
+	my $editor = (-x $ENV{SVN_EDITOR} && $ENV{SVN_EDITOR}) || (-x $ENV{VISUAL} && $ENV{VISUAL}) || (-x $ENV{EDITOR} && $ENV{EDITOR}) || 'vi';
+	if ($^O eq 'MSWin32') {
+		$editor = 'notepad.exe';
+	}
+	# need to unlock for external editor
+	flock $tfh, LOCK_UN;
+
+	my $status = system $editor, $filename;
+
+	if ( $status != 0 ) {
+		die ( $status != -1 ) ? 
+			"external editor failed: editor=$editor, errno=$?"
+		     : "could not launch program: editor=$editor, errno=$!";
+	}
+
+	unless ( seek $tfh, 0, 0 ) {
+		die "could not seek on temp file: errno=$!";
+	}
+	
+	$$msg = join '', <$tfh>;
+	$$msg =~ s/$info.*//s;
+	$log_message = $$msg;
+
+	$$msg =~ s/\r//g;
+	my  $converter = Text::Iconv->new("WINDOWS-1251", "UTF8");	
+	$$msg = $converter -> convert($$msg);
+	
+#	return;
+}
+
+
 ################################################################################
 
 sub random_password {
@@ -957,7 +1387,11 @@ sub random_password {
 
 package Zanas::Install;
 
-$VERSION = 0.55;
+$VERSION = 0.60;
+
+BEGIN {
+    $SIG {__DIE__} = \&Carp::confess;
+}
 
 =head1 NAME
 
